from __future__ import annotations

from typing import TYPE_CHECKING

import xarray as xr

from glotaran.examples.sequential import dataset
from glotaran.examples.sequential import model
from glotaran.examples.sequential import parameter
from glotaran.io import load_scheme
from glotaran.io import save_dataset
from glotaran.io import save_model
from glotaran.io import save_parameters
from glotaran.io import save_scheme
from glotaran.project import Scheme

if TYPE_CHECKING:
    from pathlib import Path


want = """add_svd: true
clp_link_tolerance: 0.0
data_files:
  dataset_1: d.nc
ftol: 1.0e-08
gtol: 1.0e-08
maximum_number_function_evaluations: null
model_file: m.yml
optimization_method: TrustRegionReflection
parameters_file: p.csv
result_path: null
xtol: 1.0e-08
"""


def test_save_scheme(tmp_path: Path):
    scheme = Scheme(
        model,
        parameter,
        {"dataset_1": dataset},
        model_file="m.yml",
        parameters_file="p.csv",
        data_files={"dataset_1": "d.nc"},
    )
    save_model(model, tmp_path / "m.yml")
    save_parameters(parameter, tmp_path / "p.csv")
    save_dataset(dataset, tmp_path / "d.nc")
    scheme_path = tmp_path / "testscheme.yml"
    save_scheme(file_name=scheme_path, format_name="yml", scheme=scheme)

    assert scheme_path.is_file()
    assert scheme_path.read_text() == want
    loaded = load_scheme(scheme_path)
    print(loaded.model.validate(loaded.parameters))
    assert loaded.model.valid(loaded.parameters)
    assert isinstance(scheme.data["dataset_1"], xr.Dataset)



"""Deprecated attributes from ``glotaran.__init__`` which are removed."""
from __future__ import annotations

from typing import TYPE_CHECKING

from glotaran.deprecation import deprecate
from glotaran.io import load_model
from glotaran.io import load_parameters

if TYPE_CHECKING:
    from glotaran.model import Model
    from glotaran.parameter import ParameterGroup


@deprecate(
    deprecated_qual_name_usage="glotaran.read_model_from_yaml(model_yml_str)",
    new_qual_name_usage='glotaran.io.load_model(model_yml_str, format_name="yml_str")',
    to_be_removed_in_version="0.6.0",
)
def read_model_from_yaml(model_yml_str: str) -> Model:
    """Parse ``yaml`` string to :class:`Model`.

    Warning
    -------
    Deprecated use ``glotaran.io.load_model(model_yml_str, format_name="yml_str")``
    instead.

    Parameters
    ----------
    model_yml_str : str
        Model spec description in yaml.

    Returns
    -------
    Model
        Model described in ``model_yml_str``.
    """
    return load_model(model_yml_str, format_name="yml_str")


@deprecate(
    deprecated_qual_name_usage="glotaran.read_model_from_yaml_file(model_file)",
    new_qual_name_usage="glotaran.io.load_model(model_file)",
    to_be_removed_in_version="0.6.0",
)
def read_model_from_yaml_file(model_file: str) -> Model:
    """Parse ``model.yaml`` file to :class:`Model`.

    Warning
    -------
    Deprecated use ``glotaran.io.load_model(model_file)`` instead.

    Parameters
    ----------
    model_file : str
        File with model spec description as yaml.

    Returns
    -------
    Model
        Model described in ``model_file``.
    """
    return load_model(model_file)


@deprecate(
    deprecated_qual_name_usage="glotaran.read_parameters_from_csv_file(parameters_file)",
    new_qual_name_usage="glotaran.io.load_parameters(parameters_file)",
    to_be_removed_in_version="0.6.0",
)
def read_parameters_from_csv_file(parameters_file: str) -> ParameterGroup:
    """Parse ``parameters_file`` to :class:`ParameterGroup`.

    Warning
    -------
    Deprecated use ``glotaran.io.load_parameters(parameters_file)`` instead.

    Parameters
    ----------
    parameters_file : str
        File with parameters in csv.

    Returns
    -------
    ParameterGroup
        ParameterGroup described in ``parameters_file``.
    """
    return load_parameters(parameters_file)


@deprecate(
    deprecated_qual_name_usage="glotaran.read_parameters_from_yaml(parameters_yml_str)",
    new_qual_name_usage="glotaran.io.load_model(parameters_yml_str)",
    to_be_removed_in_version="0.6.0",
)
def read_parameters_from_yaml(parameters_yml_str: str) -> ParameterGroup:
    """Parse ``yaml`` string to :class:`ParameterGroup`.

    Warning
    -------
    Deprecated use ``glotaran.io.load_parameters(parameters_yml_str, format_name="yml_str")``
    instead.

    Parameters
    ----------
    parameters_yml_str : str
        PArameter spec description in yaml.

    Returns
    -------
    ParameterGroup
        ParameterGroup described in ``parameters_yml_str``.
    """
    return load_parameters(parameters_yml_str, format_name="yml_str")


@deprecate(
    deprecated_qual_name_usage="glotaran.read_parameters_from_yaml_file(parameters_file)",
    new_qual_name_usage="glotaran.io.load_parameters(parameters_file)",
    to_be_removed_in_version="0.6.0",
)
def read_parameters_from_yaml_file(parameters_file: str) -> ParameterGroup:
    """Parse ``parameters_file`` to :class:`ParameterGroup`.

    Warning
    -------
    Deprecated use ``glotaran.io.load_parameters(parameters_file)`` instead.

    Parameters
    ----------
    parameters_file : str
        File with parameters in yaml.

    Returns
    -------
    ParameterGroup
        ParameterGroup described in ``parameters_file``.
    """
    return load_parameters(parameters_file)



"""The model item decorator."""
from __future__ import annotations

import copy
from typing import TYPE_CHECKING
from typing import Callable
from typing import List
from typing import Type

from glotaran.model.property import ModelProperty
from glotaran.model.util import wrap_func_as_method
from glotaran.parameter import Parameter

if TYPE_CHECKING:
    from typing import Any

    from glotaran.model.model import Model
    from glotaran.parameter import ParameterGroup

    Validator = Callable[
        [Type[object], Type[Model]],
        List[str],
    ]

    ValidatorParameter = Callable[
        [Type[object], Type[Model], Type[ParameterGroup]],
        List[str],
    ]


def model_item(
    properties: None | dict[str, dict[str, Any]] = None,
    has_type: bool = False,
    has_label: bool = True,
) -> Callable:
    """The `@model_item` decorator adds the given properties to the class. Further it adds
    classmethods for deserialization, validation and printing.

    By default, a `label` property is added.

    The `properties` dictionary contains the name of the properties as keys. The values must be
    either a `type` or dictionary with the following values:

    * type: a `type` (required)
    * doc: a string for documentation (optional)
    * default: a default value (optional)
    * allow_none: if `True`, the property can be set to None (optional)

    Classes with the `model_item` decorator intended to be used in glotaran models.

    Parameters
    ----------
    properties :
        A dictionary of property names and options.
    has_type :
        If true, a type property will added. Used for model attributes, which
        can have more then one type.
    has_label :
        If false no label property will be added.
    """

    if properties is None:
        properties = {}

    def decorator(cls):

        setattr(cls, "_glotaran_has_label", has_label)
        setattr(cls, "_glotaran_model_item", True)

        # store for later sanity checking
        if not hasattr(cls, "_glotaran_properties"):
            setattr(cls, "_glotaran_properties", [])
            if has_label:
                doc = f"The label of {cls.__name__} item."
                prop = ModelProperty(cls, "label", str, doc, None, False)
                setattr(cls, "label", prop)
                getattr(cls, "_glotaran_properties").append("label")
            if has_type:
                doc = f"The type string of {cls.__name__}."
                prop = ModelProperty(cls, "type", str, doc, None, True)
                setattr(cls, "type", prop)
                getattr(cls, "_glotaran_properties").append("type")

        else:
            setattr(
                cls,
                "_glotaran_properties",
                list(getattr(cls, "_glotaran_properties")),
            )

        for name, options in properties.items():
            if not isinstance(options, dict):
                options = {"type": options}
            prop = ModelProperty(
                cls,
                name,
                options.get("type"),
                options.get("doc", f"{name}"),
                options.get("default", None),
                options.get("allow_none", False),
            )
            setattr(cls, name, prop)
            if name not in getattr(cls, "_glotaran_properties"):
                getattr(cls, "_glotaran_properties").append(name)

        validators = _get_validators(cls)
        setattr(cls, "_glotaran_validators", validators)

        init = _create_init_func(cls)
        setattr(cls, "__init__", init)

        from_dict = _create_from_dict_func(cls)
        setattr(cls, "from_dict", from_dict)

        validate = _create_validation_func(cls)
        setattr(cls, "validate", validate)

        as_dict = _create_as_dict_func(cls)
        setattr(cls, "as_dict", as_dict)

        get_state = _create_get_state_func(cls)
        setattr(cls, "__getstate__", get_state)

        set_state = _create_set_state_func(cls)
        setattr(cls, "__setstate__", set_state)

        fill = _create_fill_func(cls)
        setattr(cls, "fill", fill)

        get_parameters = _create_get_parameters(cls)
        setattr(cls, "get_parameters", get_parameters)

        mprint = _create_mprint_func(cls)
        setattr(cls, "mprint", mprint)

        return cls

    return decorator


def model_item_typed(
    *,
    types: dict[str, Any],
    has_label: bool = True,
    default_type: str = None,
):
    """The model_item_typed decorator adds attributes to the class to enable
    the glotaran model parser to infer the correct class for an item when there
    are multiple variants.

    Parameters
    ----------
    types :
        A dictionary of types and options.
    has_label:
        If `False` no label property will be added.
    """

    def decorator(cls):

        setattr(cls, "_glotaran_model_item", True)
        setattr(cls, "_glotaran_model_item_typed", True)
        setattr(cls, "_glotaran_model_item_types", types)
        setattr(cls, "_glotaran_model_item_default_type", default_type)

        get_default_type = _create_get_default_type_func(cls)
        setattr(cls, "get_default_type", get_default_type)

        add_type = _create_add_type_func(cls)
        setattr(cls, "add_type", add_type)

        setattr(cls, "_glotaran_has_label", has_label)

        return cls

    return decorator


def model_item_validator(need_parameter: bool):
    """The model_item_validator marks a method of a model item as validation function"""

    def decorator(method: Validator | ValidatorParameter):
        setattr(method, "_glotaran_validator", need_parameter)
        return method

    return decorator


def _get_validators(cls):
    return {
        method: getattr(getattr(cls, method), "_glotaran_validator")
        for method in dir(cls)
        if hasattr(getattr(cls, method), "_glotaran_validator")
    }


def _create_get_default_type_func(cls):
    @classmethod
    @wrap_func_as_method(cls)
    def get_default_type(cls) -> str:
        return getattr(cls, "_glotaran_model_item_default_type")

    return get_default_type


def _create_add_type_func(cls):
    @classmethod
    @wrap_func_as_method(cls)
    def add_type(cls, type_name: str, attribute_type: type):
        getattr(cls, "_glotaran_model_item_types")[type_name] = attribute_type

    return add_type


def _create_init_func(cls):
    @classmethod
    @wrap_func_as_method(cls)
    def __init__(self):
        for attr in self._glotaran_properties:
            setattr(self, f"_{attr}", None)

    return __init__


def _create_from_dict_func(cls):
    @classmethod
    @wrap_func_as_method(cls)
    def from_dict(ncls, values: dict) -> cls:
        f"""Creates an instance of {cls.__name__} from a dictionary of values.

        Intended only for internal use.

        Parameters
        ----------
        values :
            A list of values.
        """
        item = ncls()

        for name in ncls._glotaran_properties:
            if name in values:
                value = values[name]
                prop = getattr(item.__class__, name)
                if prop.property_type == float:
                    value = float(value)
                elif prop.property_type == int:
                    value = int(value)
                setattr(item, name, value)

            elif not getattr(ncls, name).allow_none and getattr(item, name) is None:
                raise ValueError(f"Missing Property '{name}' For Item '{ncls.__name__}'")
        return item

    return from_dict


def _create_validation_func(cls):
    @wrap_func_as_method(cls)
    def validate(self, model: Model, parameters: ParameterGroup | None = None) -> list[str]:
        f"""Creates a list of parameters needed by this instance of {cls.__name__} not present in a
        set of parameters.

        Parameters
        ----------
        model :
            The model to validate.
        parameter :
            The parameter to validate.
        missing :
            A list the missing will be appended to.
        """
        problems = []
        for name in self._glotaran_properties:
            prop = getattr(self.__class__, name)
            value = getattr(self, name)
            problems += prop.validate(value, model, parameters)
        for validator, need_parameter in self._glotaran_validators.items():
            if need_parameter:
                if parameters is not None:
                    problems += getattr(self, validator)(model, parameters)
            else:
                problems += getattr(self, validator)(model)

        return problems

    return validate


def _create_as_dict_func(cls):
    @wrap_func_as_method(cls)
    def as_dict(self) -> dict:
        return {
            name: getattr(self.__class__, name).as_dict_value(getattr(self, name))
            for name in self._glotaran_properties
            if name != "label" and getattr(self, name) is not None
        }

    return as_dict


def _create_fill_func(cls):
    @wrap_func_as_method(cls)
    def fill(self, model: Model, parameters: ParameterGroup) -> cls:
        """Returns a copy of the {cls._name} instance with all members which are Parameters are
        replaced by the value of the corresponding parameter in the parameter group.

        Parameters
        ----------
        model :
            A glotaran model.
        parameter : ParameterGroup
            The parameter group to fill from.
        """
        item = copy.deepcopy(self)
        for name in self._glotaran_properties:
            prop = getattr(self.__class__, name)
            value = getattr(self, name)
            value = prop.fill(value, model, parameters)
            setattr(item, name, value)
        return item

    return fill


def _create_get_parameters(cls):
    @wrap_func_as_method(cls)
    def get_parameters(self) -> list[str]:
        """Returns all parameter full labels of the item."""
        parameters = []
        for name in self._glotaran_properties:
            value = getattr(self, name)
            prop = getattr(self.__class__, name)
            parameters += prop.get_parameters(value)
        return parameters

    return get_parameters


def _create_get_state_func(cls):
    @wrap_func_as_method(cls)
    def get_state(self) -> cls:
        return tuple(getattr(self, name) for name in self._glotaran_properties)

    return get_state


def _create_set_state_func(cls):
    @wrap_func_as_method(cls)
    def set_state(self, state) -> cls:
        for i, name in enumerate(self._glotaran_properties):
            setattr(self, name, state[i])

    return set_state


def _create_mprint_func(cls):
    @wrap_func_as_method(cls, name="mprint")
    def mprint_item(
        self, parameters: ParameterGroup = None, initial_parameters: ParameterGroup = None
    ) -> str:
        f"""Returns a string with the {cls.__name__} formatted in markdown."""

        s = "\n"
        if self._glotaran_has_label:
            s = f"**{self.label}**"

            if hasattr(self, "type"):
                s += f" ({self.type})"
            s += ":\n"
        elif hasattr(self, "type"):
            s = f"**{self.type}**:\n"

        attrs = []
        for name in self._glotaran_properties:
            value = getattr(self, name)
            if value is None:
                continue
            a = f"* *{name.replace('_', ' ').title()}*: "

            def format_parameter(param):
                s = f"{param.full_label}"
                if parameters is not None:
                    p = parameters.get(param.full_label)
                    s += f": **{p.value:.5e}**"
                    if p.vary:
                        err = p.standard_error or 0
                        s += f" *(StdErr: {err:.0e}"
                        if initial_parameters is not None:
                            i = initial_parameters.get(param.full_label)
                            s += f" ,initial: {i.value:.5e}"
                        s += ")*"
                    else:
                        s += " *(fixed)*"
                return s

            if isinstance(value, Parameter):
                a += format_parameter(value)
            elif isinstance(value, list) and all([isinstance(v, Parameter) for v in value]):
                a += f"[{', '.join([format_parameter(v) for v in value])}]"
            elif isinstance(value, dict):
                a += "\n"
                for k, v in value.items():
                    a += f"  * *{k}*: "
                    if isinstance(v, Parameter):
                        a += format_parameter(v)
                    else:
                        a += f"{v}"
                    a += "\n"
            else:
                a += f"{value}"
            attrs.append(a)
        s += "\n".join(attrs)
        return s

    return mprint_item



from math import inf
from math import nan
from typing import Dict
from typing import List
from typing import Tuple

import pytest
import xarray as xr
from IPython.core.formatters import format_display_data

from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import megacomplex
from glotaran.model import model_item
from glotaran.model.clp_penalties import EqualAreaPenalty
from glotaran.model.constraint import Constraint
from glotaran.model.constraint import OnlyConstraint
from glotaran.model.constraint import ZeroConstraint
from glotaran.model.interval_property import IntervalProperty
from glotaran.model.model import Model
from glotaran.model.relation import Relation
from glotaran.model.weight import Weight
from glotaran.parameter import Parameter
from glotaran.parameter import ParameterGroup


@model_item(
    properties={
        "param": Parameter,
        "megacomplex": str,
        "param_list": List[Parameter],
        "default_item": {"type": int, "default": 42},
        "complex": {"type": Dict[Tuple[str, str], Parameter]},
    },
)
class MockItem:
    pass


@model_item(
    properties={
        "param": Parameter,
        "param_list": List[Parameter],
        "param_dict": {"type": Dict[Tuple[str, str], Parameter]},
        "number": int,
    },
)
class MockItemSimple:
    pass


@model_item(has_label=False)
class MockItemNoLabel:
    pass


@megacomplex(dimension="model", model_items={"test_item1": {"type": MockItem, "allow_none": True}})
class MockMegacomplex1(Megacomplex):
    pass


@megacomplex(dimension="model", model_items={"test_item2": MockItemNoLabel})
class MockMegacomplex2(Megacomplex):
    pass


@megacomplex(model_items={"test_item3": List[MockItem]})
class MockMegacomplex3(Megacomplex):
    pass


@megacomplex(dimension="model", model_items={"test_item4": Dict[str, MockItem]})
class MockMegacomplex4(Megacomplex):
    pass


@megacomplex(
    dimension="model",
    dataset_model_items={"test_item_dataset": MockItem},
    dataset_properties={
        "test_property_dataset1": int,
        "test_property_dataset2": {"type": Parameter},
    },
)
class MockMegacomplex5(Megacomplex):
    pass


@megacomplex(dimension="model", unique=True)
class MockMegacomplex6(Megacomplex):
    pass


@megacomplex(dimension="model", model_items={"test_item_simple": MockItemSimple})
class MockMegacomplex7(Megacomplex):
    pass


@pytest.fixture
def test_model_dict():
    model_dict = {
        "megacomplex": {
            "m1": {"test_item1": "t2"},
            "m2": {"type": "type5", "dimension": "model2"},
        },
        "dataset_groups": {
            "testgroup": {"residual_function": "non_negative_least_squares", "link_clp": True}
        },
        "weights": [
            {
                "datasets": ["d1", "d2"],
                "global_interval": (1, 4),
                "model_interval": (2, 3),
                "value": 5.4,
            }
        ],
        "test_item1": {
            "t1": {
                "param": "foo",
                "megacomplex": "m1",
                "param_list": ["bar", "baz"],
                "complex": {("s1", "s2"): "baz"},
            },
            "t2": {
                "param": "baz",
                "megacomplex": "m2",
                "param_list": ["foo"],
                "complex": {},
                "default_item": 7,
            },
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ["m1"],
                "scale": "scale_1",
                "test_item_dataset": "t1",
                "test_property_dataset1": 1,
                "test_property_dataset2": "bar",
            },
            "dataset2": {
                "megacomplex": ["m2"],
                "global_megacomplex": ["m1"],
                "scale": "scale_2",
                "test_item_dataset": "t2",
                "test_property_dataset1": 1,
                "test_property_dataset2": "bar",
                "group": "testgroup",
            },
        },
    }
    model_dict["test_item_dataset"] = model_dict["test_item1"]
    return model_dict


@pytest.fixture
def test_model(test_model_dict):
    return Model.from_dict(
        test_model_dict,
        megacomplex_types={
            "type1": MockMegacomplex1,
            "type5": MockMegacomplex5,
        },
    )


@pytest.fixture
def model_error():
    model_dict = {
        "megacomplex": {"m1": {}, "m2": {"type": "type2"}, "m3": {"type": "type2"}},
        "test_item1": {
            "t1": {
                "param": "fool",
                "megacomplex": "mX",
                "param_list": ["bar", "bay"],
                "complex": {("s1", "s3"): "boz"},
            },
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ["N1", "N2"],
                "scale": "scale_1",
            },
            "dataset2": {
                "megacomplex": ["mrX"],
                "scale": "scale_3",
            },
            "dataset3": {
                "megacomplex": ["m2", "m3"],
            },
        },
    }
    return Model.from_dict(
        model_dict,
        megacomplex_types={
            "type1": MockMegacomplex1,
            "type2": MockMegacomplex6,
        },
    )


def test_model_init():
    model = Model(
        megacomplex_types={
            "type1": MockMegacomplex1,
            "type2": MockMegacomplex2,
            "type3": MockMegacomplex3,
            "type4": MockMegacomplex4,
            "type5": MockMegacomplex5,
        }
    )

    assert model.default_megacomplex == "type1"

    assert len(model.megacomplex_types) == 5
    assert "type1" in model.megacomplex_types
    assert model.megacomplex_types["type1"] == MockMegacomplex1
    assert "type2" in model.megacomplex_types
    assert model.megacomplex_types["type2"] == MockMegacomplex2

    assert hasattr(model, "test_item1")
    assert isinstance(model.test_item1, dict)
    assert "test_item1" in model._model_items
    assert issubclass(model._model_items["test_item1"], MockItem)

    assert hasattr(model, "test_item2")
    assert isinstance(model.test_item2, list)
    assert "test_item2" in model._model_items
    assert issubclass(model._model_items["test_item2"], MockItemNoLabel)

    assert hasattr(model, "test_item3")
    assert isinstance(model.test_item3, dict)
    assert "test_item3" in model._model_items
    assert issubclass(model._model_items["test_item3"], MockItem)

    assert hasattr(model, "test_item4")
    assert isinstance(model.test_item4, dict)
    assert "test_item4" in model._model_items
    assert issubclass(model._model_items["test_item4"], MockItem)

    assert hasattr(model, "test_item_dataset")
    assert isinstance(model.test_item_dataset, dict)
    assert "test_item_dataset" in model._model_items
    assert issubclass(model._model_items["test_item_dataset"], MockItem)
    assert "test_item_dataset" in model._dataset_properties
    assert issubclass(model._dataset_properties["test_item_dataset"]["type"], str)
    assert "test_property_dataset1" in model._dataset_properties
    assert issubclass(model._dataset_properties["test_property_dataset1"], int)
    assert "test_property_dataset2" in model._dataset_properties
    assert issubclass(model._dataset_properties["test_property_dataset2"]["type"], Parameter)

    assert hasattr(model, "clp_area_penalties")
    assert isinstance(model.clp_area_penalties, list)
    assert "clp_area_penalties" in model._model_items
    assert issubclass(model._model_items["clp_area_penalties"], EqualAreaPenalty)

    assert hasattr(model, "clp_constraints")
    assert isinstance(model.clp_constraints, list)
    assert "clp_constraints" in model._model_items
    assert issubclass(model._model_items["clp_constraints"], Constraint)

    assert hasattr(model, "clp_relations")
    assert isinstance(model.clp_relations, list)
    assert "clp_relations" in model._model_items
    assert issubclass(model._model_items["clp_relations"], Relation)

    assert hasattr(model, "weights")
    assert isinstance(model.weights, list)
    assert "weights" in model._model_items
    assert issubclass(model._model_items["weights"], Weight)

    assert hasattr(model, "dataset")
    assert isinstance(model.dataset, dict)
    assert "dataset" in model._model_items
    assert issubclass(model._model_items["dataset"], DatasetModel)


@pytest.fixture
def parameter():
    params = [1, 2, ["foo", 3], ["bar", 4], ["baz", 2], ["scale_1", 2], ["scale_2", 8], 4e2]
    return ParameterGroup.from_list(params)


def test_model_misc(test_model: Model):
    assert isinstance(test_model.megacomplex["m1"], MockMegacomplex1)
    assert isinstance(test_model.megacomplex["m2"], MockMegacomplex5)
    assert test_model.megacomplex["m1"].dimension == "model"
    assert test_model.megacomplex["m2"].dimension == "model2"


def test_dataset_group_models(test_model: Model):
    groups = test_model.dataset_group_models
    assert "default" in groups
    assert groups["default"].residual_function == "variable_projection"
    assert groups["default"].link_clp is None
    assert "testgroup" in groups
    assert groups["testgroup"].residual_function == "non_negative_least_squares"
    assert groups["testgroup"].link_clp


def test_dataset_groups(test_model: Model):
    groups = test_model.get_dataset_groups()
    assert "default" in groups
    assert groups["default"].model.residual_function == "variable_projection"
    assert groups["default"].model.link_clp is None
    assert "dataset1" in groups["default"].dataset_models
    assert "testgroup" in groups
    assert groups["testgroup"].model.residual_function == "non_negative_least_squares"
    assert groups["testgroup"].model.link_clp
    assert "dataset2" in groups["testgroup"].dataset_models


def test_model_validity(test_model: Model, model_error: Model, parameter: ParameterGroup):
    print(test_model.test_item1["t1"])
    print(test_model.problem_list())
    print(test_model.problem_list(parameter))
    assert test_model.valid()
    assert test_model.valid(parameter)
    print(model_error.problem_list())
    print(model_error.problem_list(parameter))
    assert not model_error.valid()
    assert len(model_error.problem_list()) == 5
    assert not model_error.valid(parameter)
    assert len(model_error.problem_list(parameter)) == 9


def test_items(test_model: Model):

    assert "m1" in test_model.megacomplex
    assert "m2" in test_model.megacomplex

    assert "t1" in test_model.test_item1
    t = test_model.test_item1.get("t1")
    assert t.param.full_label == "foo"
    assert t.megacomplex == "m1"
    assert [p.full_label for p in t.param_list] == ["bar", "baz"]
    assert t.default_item == 42
    assert ("s1", "s2") in t.complex
    assert t.complex[("s1", "s2")].full_label == "baz"
    assert "t2" in test_model.test_item1
    t = test_model.test_item1.get("t2")
    assert t.param.full_label == "baz"
    assert t.megacomplex == "m2"
    assert [p.full_label for p in t.param_list] == ["foo"]
    assert t.default_item == 7
    assert t.complex == {}

    assert "dataset1" in test_model.dataset
    assert test_model.dataset.get("dataset1").megacomplex == ["m1"]
    assert test_model.dataset.get("dataset1").scale.full_label == "scale_1"

    assert "dataset2" in test_model.dataset
    assert test_model.dataset.get("dataset2").megacomplex == ["m2"]
    assert test_model.dataset.get("dataset2").global_megacomplex == ["m1"]
    assert test_model.dataset.get("dataset2").scale.full_label == "scale_2"

    assert len(test_model.weights) == 1
    w = test_model.weights[0]
    assert w.datasets == ["d1", "d2"]
    assert w.global_interval == (1, 4)
    assert w.model_interval == (2, 3)
    assert w.value == 5.4


def test_fill(test_model: Model, parameter: ParameterGroup):
    data = xr.DataArray([[1]], coords=(("global", [0]), ("model", [0]))).to_dataset(name="data")
    dataset = test_model.dataset.get("dataset1").fill(test_model, parameter)
    dataset.set_data(data)
    assert [cmplx.label for cmplx in dataset.megacomplex] == ["m1"]
    assert dataset.scale == 2

    assert dataset.get_model_dimension() == "model"
    assert dataset.get_global_dimension() == "global"
    dataset.swap_dimensions()
    assert dataset.get_model_dimension() == "global"
    assert dataset.get_global_dimension() == "model"
    dataset.swap_dimensions()
    assert dataset.get_model_dimension() == "model"
    assert dataset.get_global_dimension() == "global"

    assert not dataset.has_global_model()

    dataset = test_model.dataset.get("dataset2").fill(test_model, parameter)
    assert [cmplx.label for cmplx in dataset.megacomplex] == ["m2"]
    assert dataset.scale == 8
    assert dataset.get_model_dimension() == "model2"
    assert dataset.get_global_dimension() == "model"

    assert dataset.has_global_model()
    assert [cmplx.label for cmplx in dataset.global_megacomplex] == ["m1"]

    t = test_model.test_item1.get("t1").fill(test_model, parameter)
    assert t.param == 3
    assert t.megacomplex.label == "m1"
    assert t.param_list == [4, 2]
    assert t.default_item == 42
    assert t.complex == {("s1", "s2"): 2}
    t = test_model.test_item1.get("t2").fill(test_model, parameter)
    assert t.param == 2
    assert t.megacomplex.label == "m2"
    assert t.param_list == [3]
    assert t.default_item == 7
    assert t.complex == {}


def test_model_as_dict():
    model_dict = {
        "default_megacomplex": "type7",
        "megacomplex": {
            "m1": {"test_item_simple": "t2", "dimension": "model"},
        },
        "test_item_simple": {
            "t1": {
                "param": "foo",
                "param_list": ["bar", "baz"],
                "param_dict": {("s1", "s2"): "baz"},
                "number": 21,
            },
        },
        "dataset_groups": {
            "default": {"link_clp": None, "residual_function": "variable_projection"}
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ["m1"],
                "scale": "scale_1",
                "group": "default",
            },
        },
    }
    model = Model.from_dict(
        model_dict,
        megacomplex_types={
            "type7": MockMegacomplex7,
        },
    )
    as_model_dict = model.as_dict()
    assert as_model_dict == model_dict


def test_model_markdown_base_heading_level(test_model: Model):
    """base_heading_level applies to all sections."""
    assert test_model.markdown().startswith("# Model")
    assert "## Test" in test_model.markdown()
    assert test_model.markdown(base_heading_level=3).startswith("### Model")
    assert "#### Test" in test_model.markdown(base_heading_level=3)


def test_model_ipython_rendering(test_model: Model):
    """Autorendering in ipython"""
    rendered_obj = format_display_data(test_model)[0]

    assert "text/markdown" in rendered_obj
    assert rendered_obj["text/markdown"].startswith("### Model")

    rendered_markdown_return = format_display_data(test_model.markdown())[0]

    assert "text/markdown" in rendered_markdown_return
    assert rendered_markdown_return["text/markdown"].startswith("# Model")


def test_interval_property():
    ip1 = IntervalProperty.from_dict({"interval": [[1, 1000]]})
    assert all(ip1.applies(x) for x in (1, 500, 100))
    assert all(not ip1.applies(x) for x in (9999, inf, nan))


def test_zero_constraint():
    zc1 = ZeroConstraint.from_dict({"interval": [[1, 400], [600, 1000]], "target": "s1"})
    assert all(zc1.applies(x) for x in (1, 2, 400, 600, 1000))
    assert all(not zc1.applies(x) for x in (400.01, 500, 599.99, 9999, inf, nan))
    assert zc1.target == "s1"
    zc2 = ZeroConstraint.from_dict({"interval": [[600, 700]], "target": "s2"})
    assert all(zc2.applies(x) for x in range(600, 700, 50))
    assert all(not zc2.applies(x) for x in (599.9999, 700.0001))
    assert zc2.target == "s2"


def test_only_constraint():
    oc1 = OnlyConstraint.from_dict({"interval": [[1, 400], (600, 1000)], "target": "spectra1"})
    assert all(oc1.applies(x) for x in (400.01, 500, 599.99, 9999, inf))
    assert all(not oc1.applies(x) for x in (1, 400, 600, 1000))
    assert oc1.target == "spectra1"
    oc2 = OnlyConstraint.from_dict({"interval": [(600, 700)], "target": "spectra2"})
    assert oc2.applies(599)
    assert not oc2.applies(650)
    assert oc2.applies(701)
    assert oc2.target == "spectra2"



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING
from typing import TypeVar

import numpy as np
import xarray as xr

from glotaran.analysis.nnls import residual_nnls
from glotaran.analysis.optimization_group_calculator import OptimizationGroupCalculator
from glotaran.analysis.optimization_group_calculator_linked import (
    OptimizationGroupCalculatorLinked,
)
from glotaran.analysis.optimization_group_calculator_unlinked import (
    OptimizationGroupCalculatorUnlinked,
)
from glotaran.analysis.util import get_min_max_from_interval
from glotaran.analysis.variable_projection import residual_variable_projection
from glotaran.io.prepare_dataset import add_svd_to_dataset
from glotaran.model import DatasetGroup
from glotaran.model import DatasetModel
from glotaran.model import Model
from glotaran.parameter import ParameterGroup
from glotaran.parameter import ParameterHistory
from glotaran.project import Scheme

if TYPE_CHECKING:
    from typing import Hashable


class InitialParameterError(ValueError):
    def __init__(self):
        super().__init__("Initial parameters can not be evaluated.")


class ParameterNotInitializedError(ValueError):
    def __init__(self):
        super().__init__("Parameter not initialized")


XrDataContainer = TypeVar("XrDataContainer", xr.DataArray, xr.Dataset)

residual_functions = {
    "variable_projection": residual_variable_projection,
    "non_negative_least_squares": residual_nnls,
}


class OptimizationGroup:
    def __init__(
        self,
        scheme: Scheme,
        dataset_group: DatasetGroup,
    ):
        """Create OptimizationGroup instance  from a scheme (:class:`glotaran.analysis.scheme.Scheme`)

        Args:
            scheme (Scheme): An instance of :class:`glotaran.analysis.scheme.Scheme`
                which defines your model, parameters, and data
        """

        self._model = scheme.model
        if scheme.parameters is None:
            raise ParameterNotInitializedError
        self._parameters = scheme.parameters.copy()
        self._dataset_group_model = dataset_group.model
        self._clp_link_tolerance = scheme.clp_link_tolerance

        try:
            self._residual_function = residual_functions[dataset_group.model.residual_function]
        except KeyError:
            raise ValueError(
                f"Unknown residual function '{dataset_group.model.residual_function}', "
                f"allowed functions are: {list(residual_functions.keys())}."
            )
        self._dataset_models = dataset_group.dataset_models

        self._overwrite_index_dependent = self.model.need_index_dependent()

        self._model.validate(raise_exception=True)

        self._prepare_data(scheme, list(dataset_group.dataset_models.keys()))
        self._dataset_labels = list(self.data.keys())

        link_clp = dataset_group.model.link_clp
        if link_clp is None:
            link_clp = self.model.is_groupable(self.parameters, self.data)

        self._calculator: OptimizationGroupCalculator = (
            OptimizationGroupCalculatorLinked(self)
            if link_clp
            else OptimizationGroupCalculatorUnlinked(self)
        )

        # all of the above are always not None

        self._matrices = None
        self._reduced_matrices = None
        self._reduced_clps = None
        self._clps = None
        self._weighted_residuals = None
        self._residuals = None
        self._additional_penalty = None
        self._full_penalty = None

    @property
    def model(self) -> Model:
        """Property providing access to the used model

        The model is a subclass of :class:`glotaran.model.Model` decorated with the `@model`
        decorator :class:`glotaran.model.model_decorator.model`
        For an example implementation see e.g. :class:`glotaran.builtin.models.kinetic_spectrum`

        Returns:
            Model: A subclass of :class:`glotaran.model.Model`
                The model must be decorated with the `@model` decorator
                :class:`glotaran.model.model_decorator.model`
        """
        return self._model

    @property
    def data(self) -> dict[str, xr.Dataset]:
        return self._data

    @property
    def parameters(self) -> ParameterGroup:
        return self._parameters

    @parameters.setter
    def parameters(self, parameters: ParameterGroup):
        self._parameters = parameters
        self.reset()

    @property
    def dataset_models(self) -> dict[str, DatasetModel]:
        return self._dataset_models

    @property
    def matrices(
        self,
    ) -> dict[str, np.ndarray | list[np.ndarray]]:
        if self._matrices is None:
            self._calculator.calculate_matrices()
        return self._matrices

    @property
    def reduced_matrices(
        self,
    ) -> dict[str, np.ndarray] | dict[str, list[np.ndarray]] | list[np.ndarray]:
        if self._reduced_matrices is None:
            self._calculator.calculate_matrices()
        return self._reduced_matrices

    @property
    def reduced_clps(
        self,
    ) -> dict[str, list[np.ndarray]]:
        if self._reduced_clps is None:
            self._calculator.calculate_residual()
        return self._reduced_clps

    @property
    def clps(
        self,
    ) -> dict[str, list[np.ndarray]]:
        if self._clps is None:
            self._calculator.calculate_residual()
        return self._clps

    @property
    def weighted_residuals(
        self,
    ) -> dict[str, list[np.ndarray]]:
        if self._weighted_residuals is None:
            self._calculator.calculate_residual()
        return self._weighted_residuals

    @property
    def residuals(
        self,
    ) -> dict[str, list[np.ndarray]]:
        if self._residuals is None:
            self._calculator.calculate_residual()
        return self._residuals

    @property
    def additional_penalty(
        self,
    ) -> dict[str, list[float]]:
        if self._additional_penalty is None:
            self._calculator.calculate_residual()
        return self._additional_penalty

    @property
    def full_penalty(self) -> np.ndarray:
        if self._full_penalty is None:
            self._calculator.calculate_full_penalty()
        return self._full_penalty

    @property
    def cost(self) -> float:
        return 0.5 * np.dot(self.full_penalty, self.full_penalty)

    def reset(self):
        """Resets all results and `DatasetModels`. Use after updating parameters."""
        self._dataset_models = {
            label: dataset_model.fill(self._model, self._parameters).set_data(self.data[label])
            for label, dataset_model in self.model.dataset.items()
            if label in self._dataset_labels
        }
        if self._overwrite_index_dependent:
            for d in self._dataset_models.values():
                d.overwrite_index_dependent(self._overwrite_index_dependent)
        self._reset_results()

    def _reset_results(self):
        self._matrices = None
        self._reduced_matrices = None
        self._reduced_clps = None
        self._clps = None
        self._weighted_residuals = None
        self._residuals = None
        self._additional_penalty = None
        self._full_penalty = None

    def _prepare_data(self, scheme: Scheme, labels: list[str]):
        self._data = {}
        self._dataset_models = {}
        for label, dataset in scheme.data.items():
            if label not in labels:
                continue
            if isinstance(dataset, xr.DataArray):
                dataset = dataset.to_dataset(name="data")

            dataset_model = self._model.dataset[label]
            dataset_model = dataset_model.fill(self.model, self.parameters)
            dataset_model.set_data(dataset)
            if self._overwrite_index_dependent:
                dataset_model.overwrite_index_dependent(self._overwrite_index_dependent)
            self._dataset_models[label] = dataset_model
            global_dimension = dataset_model.get_global_dimension()
            model_dimension = dataset_model.get_model_dimension()

            dataset = self._transpose_dataset(
                dataset, ordered_dims=[model_dimension, global_dimension]
            )

            if scheme.add_svd:
                add_svd_to_dataset(dataset, lsv_dim=model_dimension, rsv_dim=global_dimension)

            self._add_weight(label, dataset)
            self._data[label] = dataset

    def _transpose_dataset(
        self, datacontainer: XrDataContainer, ordered_dims: list[Hashable]
    ) -> XrDataContainer:
        """Reorder dimension of the datacontainer with the order provided by ``ordered_dims``.

        Parameters
        ----------
        dataset: XrDataContainer
            Dataset to be reordered
        ordered_dims: list[Hashable]
            Order the dimensions should be in.

        Returns
        -------
        XrDataContainer
            Datacontainer with reordered dimensions.
        """
        ordered_dims = list(filter(lambda dim: dim in datacontainer.dims, ordered_dims))
        ordered_dims += list(filter(lambda dim: dim not in ordered_dims, datacontainer.dims))
        return datacontainer.transpose(*ordered_dims)

    def _add_weight(self, label, dataset):

        # if the user supplies a weight we ignore modeled weights
        if "weight" in dataset:
            if any(label in weight.datasets for weight in self.model.weights):
                warnings.warn(
                    f"Ignoring model weight for dataset '{label}'"
                    " because weight is already supplied by dataset."
                )
            return
        dataset_model = self.dataset_models[label]
        dataset_model.set_data(dataset)
        global_dimension = dataset_model.get_global_dimension()
        model_dimension = dataset_model.get_model_dimension()

        global_axis = dataset.coords[global_dimension]
        model_axis = dataset.coords[model_dimension]

        for weight in self.model.weights:
            if label in weight.datasets:
                if "weight" not in dataset:
                    dataset["weight"] = xr.DataArray(
                        np.ones_like(dataset.data), coords=dataset.data.coords
                    )

                idx = {}
                if weight.global_interval is not None:
                    idx[global_dimension] = get_min_max_from_interval(
                        weight.global_interval, global_axis
                    )
                if weight.model_interval is not None:
                    idx[model_dimension] = get_min_max_from_interval(
                        weight.model_interval, model_axis
                    )
                dataset.weight[idx] *= weight.value

    def create_result_data(
        self,
        parameter_history: ParameterHistory = None,
        copy: bool = True,
        success: bool = True,
        add_svd: bool = True,
    ) -> dict[str, xr.Dataset]:

        if not success:
            if parameter_history is not None and parameter_history.number_of_records > 1:
                self.parameters.set_from_history(parameter_history, -2)
            else:
                raise InitialParameterError()

        self.reset()
        self._calculator.prepare_result_creation()
        result_data = {}
        for label, dataset_model in self.dataset_models.items():
            result_data[label] = self.create_result_dataset(label, copy=copy)
            dataset_model.finalize_data(result_data[label])

        return result_data

    def create_result_dataset(
        self, label: str, copy: bool = True, add_svd: bool = True
    ) -> xr.Dataset:
        dataset = self.data[label]
        dataset_model = self.dataset_models[label]
        global_dimension = dataset_model.get_global_dimension()
        model_dimension = dataset_model.get_model_dimension()
        if copy:
            dataset = dataset.copy()
        if dataset_model.is_index_dependent():
            dataset = self._calculator.create_index_dependent_result_dataset(label, dataset)
        else:
            dataset = self._calculator.create_index_independent_result_dataset(label, dataset)

        # TODO: adapt tests to handle add_svd=False
        if add_svd:
            self._create_svd("weighted_residual", dataset, model_dimension, global_dimension)
            self._create_svd("residual", dataset, model_dimension, global_dimension)

        # Calculate RMS
        size = dataset.residual.shape[0] * dataset.residual.shape[1]
        dataset.attrs["root_mean_square_error"] = np.sqrt(
            (dataset.residual ** 2).sum() / size
        ).values
        size = dataset.weighted_residual.shape[0] * dataset.weighted_residual.shape[1]
        dataset.attrs["weighted_root_mean_square_error"] = np.sqrt(
            (dataset.weighted_residual ** 2).sum() / size
        ).values
        if dataset_model.scale is not None:
            dataset.attrs["dataset_scale"] = dataset_model.scale.value
        else:
            dataset.attrs["dataset_scale"] = 1

        # reconstruct fitted data
        dataset["fitted_data"] = dataset.data - dataset.residual
        return dataset

    def _create_svd(self, name: str, dataset: xr.Dataset, lsv_dim: str, rsv_dim: str):
        """Calculate the SVD of a data matrix in the dataset and add it to the dataset.

        Parameters
        ----------
        name : str
            Name of the data matrix.
        dataset : xr.Dataset
            Dataset containing the data, which will be updated with the SVD values.
        """
        data_array: xr.DataArray = self._transpose_dataset(
            dataset[name],
            ordered_dims=[lsv_dim, rsv_dim],
        )

        add_svd_to_dataset(
            dataset, name=name, lsv_dim=lsv_dim, rsv_dim=rsv_dim, data_array=data_array
        )



"""This package contains compartment constraint items."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import List
from typing import Tuple

import numpy as np
import xarray as xr

from glotaran.model.item import model_item
from glotaran.parameter import Parameter

if TYPE_CHECKING:
    from typing import Any
    from typing import Sequence

    from glotaran.model.model import Model
    from glotaran.parameter import ParameterGroup


@model_item(
    properties={
        "source": str,
        "source_intervals": List[Tuple[float, float]],
        "target": str,
        "target_intervals": List[Tuple[float, float]],
        "parameter": Parameter,
        "weight": str,
    },
    has_label=False,
)
class EqualAreaPenalty:
    """An equal area constraint adds a the differenc of the sum of a
    compartments in the e matrix in one ore more intervals to the scaled sum
    of the e matrix of one or more target compartments to residual. The additional
    residual is scaled with the weight."""

    def applies(self, index: Any) -> bool:
        """
        Returns true if the index is in one of the intervals.

        Parameters
        ----------
        index :

        Returns
        -------
        applies : bool

        """

        def applies(interval):
            return interval[0] <= index <= interval[1]

        if isinstance(self.interval, tuple):
            return applies(self.interval)
        return any([applies(i) for i in self.interval])


def has_spectral_penalties(model: Model) -> bool:
    return len(model.clp_area_penalties) != 0


def apply_spectral_penalties(
    model: Model,
    parameters: ParameterGroup,
    clp_labels: dict[str, list[str] | list[list[str]]],
    clps: dict[str, list[np.ndarray]],
    matrices: dict[str, np.ndarray | list[np.ndarray]],
    data: dict[str, xr.Dataset],
    group_tolerance: float,
) -> np.ndarray:

    # TODO: seems to duplicate calculate_clp_penalties
    penalties = []
    for penalty in model.clp_area_penalties:

        penalty = penalty.fill(model, parameters)
        source_area = _get_area(
            model.index_dependent(),
            model.global_dimension,
            clp_labels,
            clps,
            data,
            group_tolerance,
            penalty.source_intervals,
            penalty.source,
        )

        target_area = _get_area(
            model.index_dependent(),
            model.global_dimension,
            clp_labels,
            clps,
            data,
            group_tolerance,
            penalty.target_intervals,
            penalty.target,
        )

        area_penalty = np.abs(np.sum(source_area) - penalty.parameter * np.sum(target_area))
        penalties.append(area_penalty * penalty.weight)
    return np.asarray(penalties)


def _get_area(
    index_dependent: bool,
    global_dimension: str,
    clp_labels: dict[str, list[list[str]]],
    clps: dict[str, list[np.ndarray]],
    data: dict[str, xr.Dataset],
    group_tolerance: float,
    intervals: list[tuple[float, float]],
    compartment: str,
) -> np.ndarray:
    area = []
    area_indices = []

    for label, dataset in data.items():
        global_axis = dataset.coords[global_dimension]
        for interval in intervals:
            if interval[0] > global_axis[-1]:
                # interval not in this dataset
                continue

            start_idx, end_idx = _get_idx_from_interval(interval, global_axis)
            for i in range(start_idx, end_idx + 1):
                index_clp_labels = clp_labels[label][i] if index_dependent else clp_labels[label]
                if compartment in index_clp_labels:
                    area.append(clps[label][i][index_clp_labels.index(compartment)])
                    area_indices.append(global_axis[i])

    return np.asarray(area)  # TODO: normalize for distance on global axis


def _get_idx_from_interval(
    interval: tuple[float, float], axis: Sequence[float] | np.ndarray
) -> tuple[int, int]:
    """Retrieves start and end index of an interval on some axis

    Parameters
    ----------
    interval : A tuple of floats with begin and end of the interval
    axis : Array like object which can be cast to np.array

    Returns
    -------
    start, end : tuple of int

    """
    axis_array = np.array(axis)
    start = np.abs(axis_array - interval[0]).argmin() if not np.isinf(interval[0]) else 0
    end = (
        np.abs(axis_array - interval[1]).argmin()
        if not np.isinf(interval[1])
        else axis_array.size - 1
    )
    return start, end



"""Glotaran package __init__.py"""
from glotaran.deprecation.modules.glotaran_root import read_model_from_yaml
from glotaran.deprecation.modules.glotaran_root import read_model_from_yaml_file
from glotaran.deprecation.modules.glotaran_root import read_parameters_from_csv_file
from glotaran.deprecation.modules.glotaran_root import read_parameters_from_yaml
from glotaran.deprecation.modules.glotaran_root import read_parameters_from_yaml_file
from glotaran.plugin_system.base_registry import load_plugins

load_plugins()

__version__ = "0.5.0rc1"


def __getattr__(attribute_name: str):
    from glotaran.deprecation.deprecation_utils import deprecate_module_attribute

    if attribute_name == "ParameterGroup":
        return deprecate_module_attribute(
            deprecated_qual_name="glotaran.ParameterGroup",
            new_qual_name="glotaran.parameter.ParameterGroup",
            to_be_removed_in_version="0.6.0",
        )

    raise AttributeError(f"module {__name__} has no attribute {attribute_name}")



import numpy as np
import pytest
import xarray as xr

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.analysis.test.models import FullModel
from glotaran.analysis.test.models import MultichannelMulticomponentDecay
from glotaran.analysis.test.models import OneCompartmentDecay
from glotaran.analysis.test.models import ThreeDatasetDecay
from glotaran.analysis.test.models import TwoCompartmentDecay
from glotaran.project import Scheme


@pytest.mark.parametrize("is_index_dependent", [True, False])
@pytest.mark.parametrize("link_clp", [True, False])
@pytest.mark.parametrize("weight", [True, False])
@pytest.mark.parametrize(
    "method",
    [
        "TrustRegionReflection",
        "Dogbox",
        "Levenberg-Marquardt",
    ],
)
@pytest.mark.parametrize(
    "suite",
    [OneCompartmentDecay, TwoCompartmentDecay, ThreeDatasetDecay, MultichannelMulticomponentDecay],
)
def test_optimization(suite, is_index_dependent, link_clp, weight, method):
    model = suite.model

    model.megacomplex["m1"].is_index_dependent = is_index_dependent

    print("Link CLP:", link_clp)
    print("Index dependent:", is_index_dependent)

    sim_model = suite.sim_model
    sim_model.megacomplex["m1"].is_index_dependent = is_index_dependent

    print(model.validate())
    assert model.valid()

    print(sim_model.validate())
    assert sim_model.valid()

    wanted_parameters = suite.wanted_parameters
    print(wanted_parameters)
    print(sim_model.validate(wanted_parameters))
    assert sim_model.valid(wanted_parameters)

    initial_parameters = suite.initial_parameters
    print(initial_parameters)
    print(model.validate(initial_parameters))
    assert model.valid(initial_parameters)
    assert (
        model.dataset["dataset1"].fill(model, initial_parameters).is_index_dependent()
        == is_index_dependent
    )

    nr_datasets = 3 if issubclass(suite, ThreeDatasetDecay) else 1
    data = {}
    for i in range(nr_datasets):
        global_axis = getattr(suite, "global_axis" if i == 0 else f"global_axis{i+1}")
        model_axis = getattr(suite, "model_axis" if i == 0 else f"model_axis{i+1}")

        dataset = simulate(
            sim_model,
            f"dataset{i+1}",
            wanted_parameters,
            {"global": global_axis, "model": model_axis},
        )
        print(f"Dataset {i+1}")
        print("=============")
        print(dataset)

        if hasattr(suite, "scale"):
            dataset["data"] /= suite.scale

        if weight:
            dataset["weight"] = xr.DataArray(
                np.ones_like(dataset.data) * 0.5, coords=dataset.data.coords
            )

        assert dataset.data.shape == (model_axis.size, global_axis.size)

        data[f"dataset{i+1}"] = dataset

    scheme = Scheme(
        model=model,
        parameters=initial_parameters,
        data=data,
        maximum_number_function_evaluations=10,
        clp_link_tolerance=0.1,
        optimization_method=method,
    )

    model.dataset_group_models["default"].link_clp = link_clp

    result = optimize(scheme, raise_exception=True)
    print(result.optimized_parameters)
    assert result.success
    optimized_scheme = result.get_scheme()
    assert result.optimized_parameters == optimized_scheme.parameters
    for dataset in optimized_scheme.data.values():
        assert "fitted_data" not in dataset
        if weight:
            assert "weight" in dataset
    for label, param in result.optimized_parameters.all():
        if param.vary:
            assert np.allclose(param.value, wanted_parameters.get(label).value, rtol=1e-1)

    for i, dataset in enumerate(data.values()):
        resultdata = result.data[f"dataset{i+1}"]
        print(f"Result Data {i+1}")
        print("=================")
        print(resultdata)
        assert "residual" in resultdata
        assert "residual_left_singular_vectors" in resultdata
        assert "residual_right_singular_vectors" in resultdata
        assert "residual_singular_values" in resultdata
        assert np.array_equal(dataset.coords["model"], resultdata.coords["model"])
        assert np.array_equal(dataset.coords["global"], resultdata.coords["global"])
        assert dataset.data.shape == resultdata.data.shape
        print(dataset.data[0, 0], resultdata.data[0, 0])
        assert np.allclose(dataset.data, resultdata.data)
        if weight:
            assert "weight" in resultdata
            assert "weighted_residual" in resultdata
            assert "weighted_residual_left_singular_vectors" in resultdata
            assert "weighted_residual_right_singular_vectors" in resultdata
            assert "weighted_residual_singular_values" in resultdata


@pytest.mark.parametrize("index_dependent", [True, False])
def test_optimization_full_model(index_dependent):
    model = FullModel.model
    model.megacomplex["m1"].is_index_dependent = index_dependent

    print(model.validate())
    assert model.valid()

    parameters = FullModel.parameters
    assert model.valid(parameters)

    dataset = simulate(model, "dataset1", parameters, FullModel.coordinates)

    scheme = Scheme(
        model=model,
        parameters=parameters,
        data={"dataset1": dataset},
        maximum_number_function_evaluations=10,
    )

    result = optimize(scheme, raise_exception=True)
    assert result.success
    optimized_scheme = result.get_scheme()
    assert result.optimized_parameters == optimized_scheme.parameters

    result_data = result.data["dataset1"]
    assert "fitted_data" in result_data
    for label, param in result.optimized_parameters.all():
        if param.vary:
            assert np.allclose(param.value, parameters.get(label).value, rtol=1e-1)

    clp = result_data.clp
    print(clp)
    assert clp.shape == (4, 4)
    assert all(np.isclose(1.0, c) for c in np.diagonal(clp))



"""The module for :class:``Scheme``."""
from __future__ import annotations

from dataclasses import dataclass
from dataclasses import fields
from typing import TYPE_CHECKING

from glotaran.deprecation import deprecate
from glotaran.deprecation import warn_deprecated
from glotaran.io import load_dataset
from glotaran.io import load_model
from glotaran.io import load_parameters
from glotaran.io import load_scheme
from glotaran.project.dataclass_helpers import exclude_from_dict_field
from glotaran.project.dataclass_helpers import file_representation_field
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:

    from typing import Literal

    import xarray as xr

    from glotaran.model import Model
    from glotaran.parameter import ParameterGroup


@dataclass
class Scheme:
    """A scheme is a collection of a model, parameters and a dataset.

    A scheme also holds options for optimization.
    """

    model: Model = exclude_from_dict_field()
    parameters: ParameterGroup = exclude_from_dict_field()
    data: dict[str, xr.DataArray | xr.Dataset] = exclude_from_dict_field()
    model_file: str | None = file_representation_field("model", load_model, default=None)
    parameters_file: str | None = file_representation_field("parameters", load_parameters, None)
    data_files: dict[str, str] | None = file_representation_field("data", load_dataset, None)
    clp_link_tolerance: float = 0.0
    maximum_number_function_evaluations: int | None = None
    non_negative_least_squares: bool | None = exclude_from_dict_field(None)
    group_tolerance: float | None = exclude_from_dict_field(None)
    group: bool | None = exclude_from_dict_field(None)
    add_svd: bool = True
    ftol: float = 1e-8
    gtol: float = 1e-8
    xtol: float = 1e-8
    optimization_method: Literal[
        "TrustRegionReflection",
        "Dogbox",
        "Levenberg-Marquardt",
    ] = "TrustRegionReflection"
    result_path: str | None = None

    def __post_init__(self):
        """Override attributes after initialization."""
        if self.non_negative_least_squares is not None:
            warn_deprecated(
                deprecated_qual_name_usage=(
                    "glotaran.project.Scheme(..., non_negative_least_squares=...)"
                ),
                new_qual_name_usage="<model_file>dataset_groups.default.residual_function",
                to_be_removed_in_version="0.7.0",
                check_qual_names=(True, False),
                stacklevel=4,
            )

            default_group = self.model.dataset_group_models["default"]
            if self.non_negative_least_squares is True:
                default_group.residual_function = "non_negative_least_squares"
            else:
                default_group.residual_function = "variable_projection"
            for field in fields(self):
                if field.name == "non_negative_least_squares":
                    field.metadata = {}

        if self.group is not None:
            warn_deprecated(
                deprecated_qual_name_usage="glotaran.project.Scheme(..., group=...)",
                new_qual_name_usage="<model_file>dataset_groups.default.link_clp",
                to_be_removed_in_version="0.7.0",
                check_qual_names=(True, False),
                stacklevel=4,
            )
            self.model.dataset_group_models["default"].link_clp = self.group
            for field in fields(self):
                if field.name == "group":
                    field.metadata = {}

        if self.group_tolerance is not None:
            warn_deprecated(
                deprecated_qual_name_usage="glotaran.project.Scheme(..., group_tolerance=...)",
                new_qual_name_usage="glotaran.project.Scheme(..., clp_link_tolerance=...)",
                to_be_removed_in_version="0.7.0",
                stacklevel=4,
            )
            self.clp_link_tolerance = self.group_tolerance

    def problem_list(self) -> list[str]:
        """Return a list with all problems in the model and missing parameters.

        Returns
        -------
        list[str]
            A list of all problems found in the scheme's model.
        """
        return self.model.problem_list(self.parameters)

    def validate(self) -> str:
        """Return a string listing all problems in the model and missing parameters.

        Returns
        -------
        str
            A user-friendly string containing all the problems of a model if any.
            Defaults to 'Your model is valid.' if no problems are found.
        """
        return self.model.validate(self.parameters)

    def valid(self) -> bool:
        """Check if there are no problems with the model or the parameters.

        Returns
        -------
        bool
            Whether the scheme is valid.
        """
        return self.model.valid(self.parameters)

    def markdown(self):
        """Format the :class:`Scheme` as markdown string.

        Returns
        -------
        MarkdownStr
            The scheme as markdown string.
        """
        model_markdown_str = self.model.markdown(parameters=self.parameters)

        markdown_str = "\n\n"
        markdown_str += "__Scheme__\n\n"

        if self.non_negative_least_squares is not None:
            markdown_str += f"* *non_negative_least_squares*: {self.non_negative_least_squares}\n"
        markdown_str += (
            "* *maximum_number_function_evaluations*: "
            f"{self.maximum_number_function_evaluations}\n"
        )
        markdown_str += f"* *clp_link_tolerance*: {self.clp_link_tolerance}\n"

        return model_markdown_str + MarkdownStr(markdown_str)

    def _repr_markdown_(self) -> str:
        """Return a markdown representation str.

        Special method used by ``ipython`` to render markdown.

        Returns
        -------
        str
            The scheme as markdown string.
        """
        return str(self.markdown())

    def __str__(self) -> str:
        """Representation used by print and str."""
        return str(self.markdown())

    @property
    def model_dimensions(self) -> dict[str, str]:
        """Return the dataset model's model dimension.

        Returns
        -------
        dict[str, str]
            A dictionary with the dataset labels as key and the model dimension of
            the dataset as value.
        """
        return {
            dataset_name: self.model.dataset[dataset_name]  # type:ignore[attr-defined]
            .fill(self.model, self.parameters)
            .set_data(self.data[dataset_name])
            .get_model_dimension()
            for dataset_name in self.data
        }

    @property
    def global_dimensions(self) -> dict[str, str]:
        """Return the dataset model's global dimension.

        Returns
        -------
        dict[str, str]
            A dictionary with the dataset labels as key and the global dimension of
            the dataset as value.
        """
        return {
            dataset_name: self.model.dataset[dataset_name]  # type:ignore[attr-defined]
            .fill(self.model, self.parameters)
            .set_data(self.data[dataset_name])
            .get_global_dimension()
            for dataset_name in self.data
        }

    @staticmethod
    @deprecate(
        deprecated_qual_name_usage="glotaran.project.scheme.Scheme.from_yaml_file(filename)",
        new_qual_name_usage="glotaran.io.load_scheme(filename)",
        to_be_removed_in_version="0.6.0",
        importable_indices=(2, 1),
    )
    def from_yaml_file(filename: str) -> Scheme:
        """Create :class:`Scheme` from specs in yaml file.

        Warning
        -------
        Deprecated use ``glotaran.io.load_scheme(filename)`` instead.

        Parameters
        ----------
        filename : str
            Path to the spec file.

        Returns
        -------
        Scheme
            Analysis schmeme
        """
        return load_scheme(filename)



from __future__ import annotations

from typing import Any
from typing import NamedTuple

import pytest

from glotaran.utils.sanitize import sanitize_list_with_broken_tuples


class MangledListTestData(NamedTuple):
    input: list[Any]
    input_sanitized: list[str] | str
    output: list[str]


test_data_list = [
    MangledListTestData(
        ["(3", 100, "13)", "(4.0", "-200)"],
        "[(3, 100, 13), (4.0, -200)]",
        ["(3, 100, 13)", "(4.0, -200)"],
    ),
    MangledListTestData(
        [(3, 100, 13), 5.5, (4.0, -200), 5.6],
        "[(3, 100, 13), 5.5, (4.0, -200), 5.6]",
        ["(3, 100, 13)", "5.5", "(4.0, -200)", "5.6"],
    ),
    MangledListTestData(
        [(3, 100, 13), -5.5, (4.0, -200), +5.6],
        "[(3, 100, 13), -5.5, (4.0, -200), 5.6]",
        ["(3, 100, 13)", "-5.5", "(4.0, -200)", "5.6"],
    ),
    MangledListTestData(
        ["(3", -100, "13)", "(4.0", "-200)"],
        "[(3, -100, 13), (4.0, -200)]",
        ["(3, -100, 13)", "(4.0, -200)"],
    ),
    MangledListTestData(
        ["(3", 100, "13)", "(4.0", "+200)"],
        "[(3, 100, 13), (4.0, +200)]",
        ["(3, 100, 13)", "(4.0, +200)"],
    ),
]


@pytest.mark.parametrize("test_data", test_data_list)
def test_mangled_list_sanitization(test_data: MangledListTestData):
    assert test_data.input_sanitized == str(test_data.input).replace("'", "")


@pytest.mark.parametrize("test_data", test_data_list)
def test_fix_tuple_string_list(test_data: MangledListTestData):
    actual = sanitize_list_with_broken_tuples(test_data.input)
    assert all(a in b for a, b in zip(actual, test_data.output))


if __name__ == "__main__":
    for test_data in test_data_list:
        test_mangled_list_sanitization(test_data)
        test_fix_tuple_string_list(test_data)



import collections

import numpy as np
import pytest
import xarray as xr

from glotaran.analysis.optimization_group import OptimizationGroup
from glotaran.analysis.optimization_group_calculator_linked import (
    OptimizationGroupCalculatorLinked,
)
from glotaran.analysis.simulation import simulate
from glotaran.analysis.test.models import FullModel
from glotaran.analysis.test.models import MultichannelMulticomponentDecay as suite
from glotaran.analysis.test.models import SimpleTestModel
from glotaran.analysis.util import CalculatedMatrix
from glotaran.parameter import ParameterGroup
from glotaran.project import Scheme


@pytest.fixture(
    scope="module", params=[[True, True], [True, False], [False, True], [False, False]]
)
def optimization_group(request) -> OptimizationGroup:
    model = suite.model
    model.megacomplex["m1"].is_index_dependent = request.param[1]
    model.is_index_dependent = request.param[1]
    model.dataset_group_models["default"].link_clp = request.param[0]

    dataset = simulate(
        suite.sim_model,
        "dataset1",
        suite.wanted_parameters,
        {"global": suite.global_axis, "model": suite.model_axis},
    )
    scheme = Scheme(model=model, parameters=suite.initial_parameters, data={"dataset1": dataset})

    return OptimizationGroup(scheme, model.get_dataset_groups()["default"])


def test_optimization_group_bag(optimization_group: OptimizationGroup):

    if isinstance(optimization_group._calculator, OptimizationGroupCalculatorLinked):
        bag = optimization_group._calculator.bag
        assert isinstance(bag, collections.deque)
        assert len(bag) == suite.global_axis.size
        assert optimization_group._calculator.groups == {"dataset1": ["dataset1"]}


def test_optimization_group_matrices(optimization_group: OptimizationGroup):
    optimization_group._calculator.calculate_matrices()

    if isinstance(optimization_group._calculator, OptimizationGroupCalculatorLinked):
        if optimization_group.model.is_index_dependent:
            assert all(
                isinstance(m, CalculatedMatrix) for m in optimization_group.reduced_matrices
            )
            assert len(optimization_group.reduced_matrices) == suite.global_axis.size
        else:
            assert "dataset1" in optimization_group.reduced_matrices
            assert isinstance(optimization_group.reduced_matrices["dataset1"], CalculatedMatrix)
    else:
        if optimization_group.model.is_index_dependent:
            assert isinstance(optimization_group.reduced_matrices, dict)
            assert isinstance(optimization_group.reduced_matrices["dataset1"], list)
            assert all(
                isinstance(m, CalculatedMatrix)
                for m in optimization_group.reduced_matrices["dataset1"]
            )
        else:
            assert isinstance(optimization_group.reduced_matrices["dataset1"], CalculatedMatrix)

        assert isinstance(optimization_group.matrices, dict)
        assert "dataset1" in optimization_group.reduced_matrices


def test_optimization_group_residuals(optimization_group: OptimizationGroup):
    optimization_group._calculator.calculate_residual()
    if isinstance(optimization_group._calculator, OptimizationGroupCalculatorLinked):
        assert isinstance(optimization_group.residuals, list)
        assert all(isinstance(r, np.ndarray) for r in optimization_group.residuals)
        assert len(optimization_group.residuals) == suite.global_axis.size
    else:
        assert isinstance(optimization_group.residuals, dict)
        assert "dataset1" in optimization_group.residuals
        assert all(isinstance(r, np.ndarray) for r in optimization_group.residuals["dataset1"])
        assert len(optimization_group.residuals["dataset1"]) == suite.global_axis.size


def test_optimization_group_result_data(optimization_group: OptimizationGroup):

    data = optimization_group.create_result_data()
    label = "dataset1"

    assert label in data

    dataset = data[label]
    dataset_model = optimization_group.dataset_models[label]

    assert "clp_label" in dataset.coords
    assert np.array_equal(dataset.clp_label, ["s1", "s2", "s3", "s4"])

    assert dataset_model.get_global_dimension() in dataset.coords
    assert np.array_equal(dataset.coords[dataset_model.get_global_dimension()], suite.global_axis)

    assert dataset_model.get_model_dimension() in dataset.coords
    assert np.array_equal(dataset.coords[dataset_model.get_model_dimension()], suite.model_axis)

    assert "matrix" in dataset
    matrix = dataset.matrix
    if optimization_group.model.is_index_dependent:
        assert len(matrix.shape) == 3
        assert matrix.shape[0] == suite.global_axis.size
        assert matrix.shape[1] == suite.model_axis.size
        assert matrix.shape[2] == 4
    else:
        assert len(matrix.shape) == 2
        assert matrix.shape[0] == suite.model_axis.size
        assert matrix.shape[1] == 4

    assert "clp" in dataset
    clp = dataset.clp
    assert len(clp.shape) == 2
    assert clp.shape[0] == suite.global_axis.size
    assert clp.shape[1] == 4

    assert "weighted_residual" in dataset
    assert dataset.data.shape == dataset.weighted_residual.shape

    assert "residual" in dataset
    assert dataset.data.shape == dataset.residual.shape

    assert "residual_singular_values" in dataset
    assert "weighted_residual_singular_values" in dataset


def test_prepare_data():
    model_dict = {
        "megacomplex": {"m1": {"is_index_dependent": False}},
        "dataset": {
            "dataset1": {
                "megacomplex": ["m1"],
            },
        },
        "weights": [
            {
                "datasets": ["dataset1"],
                "global_interval": (np.inf, 200),
                "model_interval": (4, 8),
                "value": 0.5,
            },
        ],
    }
    model = SimpleTestModel.from_dict(model_dict)
    print(model.validate())
    assert model.valid()

    parameters = ParameterGroup.from_list([])

    global_axis = np.asarray(range(50, 300))
    model_axis = np.asarray(range(15))

    dataset = xr.DataArray(
        np.ones((global_axis.size, model_axis.size)),
        coords={"global": global_axis, "model": model_axis},
        dims=("global", "model"),
    )

    scheme = Scheme(model, parameters, {"dataset1": dataset})
    optimization_group = OptimizationGroup(scheme, model.get_dataset_groups()["default"])

    data = optimization_group.data["dataset1"]
    print(data)
    assert "data" in data
    assert "weight" in data

    assert data.data.shape == (model_axis.size, global_axis.size)
    assert data.data.shape == data.weight.shape
    assert np.all(data.weight.sel({"global": slice(0, 200), "model": slice(4, 8)}).values == 0.5)
    assert np.all(data.weight.sel(model=slice(0, 3)).values == 1)

    model_dict["weights"].append(
        {
            "datasets": ["dataset1"],
            "value": 0.2,
        }
    )
    model = SimpleTestModel.from_dict(model_dict)
    print(model.validate())
    assert model.valid()

    scheme = Scheme(model, parameters, {"dataset1": dataset})
    optimization_group = OptimizationGroup(scheme, model.get_dataset_groups()["default"])

    data = optimization_group.data["dataset1"]
    assert np.all(
        data.weight.sel({"global": slice(0, 200), "model": slice(4, 8)}).values == 0.5 * 0.2
    )
    assert np.all(data.weight.sel(model=slice(0, 3)).values == 0.2)

    with pytest.warns(
        UserWarning,
        match="Ignoring model weight for dataset 'dataset1'"
        " because weight is already supplied by dataset.",
    ):
        OptimizationGroup(
            Scheme(model, parameters, {"dataset1": data}), model.get_dataset_groups()["default"]
        )


def test_full_model_problem():
    dataset = simulate(FullModel.model, "dataset1", FullModel.parameters, FullModel.coordinates)
    scheme = Scheme(
        model=FullModel.model, parameters=FullModel.parameters, data={"dataset1": dataset}
    )
    optimization_group = OptimizationGroup(scheme, FullModel.model.get_dataset_groups()["default"])

    result = optimization_group.create_result_data()["dataset1"]
    assert "global_matrix" in result
    assert "global_clp_label" in result

    clp = result.clp

    assert clp.shape == (4, 4)
    print(np.diagonal(clp))
    assert all(np.isclose(1.0, c) for c in np.diagonal(clp))



from __future__ import annotations

from typing import Any
from typing import Generator
from typing import Hashable

import numpy as np
import xarray as xr

from glotaran.model.megacomplex import Megacomplex
from glotaran.model.model import Model
from glotaran.parameter import Parameter

def create_dataset_model_type(properties: dict[str, Any]) -> type[DatasetModel]: ...

class DatasetModel:

    label: str
    megacomplex: list[str]
    megacomplex_scale: list[Parameter] | None
    global_megacomplex: list[str]
    global_megacomplex_scale: list[Parameter] | None
    scale: Parameter | None
    _coords: dict[Hashable, np.ndarray]
    def iterate_megacomplexes(
        self,
    ) -> Generator[tuple[Parameter | int | None, Megacomplex | str], None, None]: ...
    def iterate_global_megacomplexes(
        self,
    ) -> Generator[tuple[Parameter | int | None, Megacomplex | str], None, None]: ...
    def get_model_dimension(self) -> str: ...
    def finalize_data(self, dataset: xr.Dataset) -> None: ...
    def overwrite_model_dimension(self, model_dimension: str) -> None: ...
    def get_global_dimension(self) -> str: ...
    def overwrite_global_dimension(self, global_dimension: str) -> None: ...
    def swap_dimensions(self) -> None: ...
    def set_data(self, dataset: xr.Dataset) -> DatasetModel: ...
    def get_data(self) -> np.ndarray: ...
    def get_weight(self) -> np.ndarray | None: ...
    def is_index_dependent(self) -> bool: ...
    def overwrite_index_dependent(self, index_dependent: bool): ...
    def has_global_model(self) -> bool: ...
    def set_coordinates(self, coords: dict[str, np.ndarray]): ...
    def get_coordinates(self) -> dict[Hashable, np.ndarray]: ...
    def get_model_axis(self) -> np.ndarray: ...
    def get_global_axis(self) -> np.ndarray: ...
    def ensure_unique_megacomplexes(self, model: Model) -> list[str]: ...



"""The result class for global analysis."""
from __future__ import annotations

from dataclasses import dataclass
from dataclasses import replace
from typing import Any
from typing import Dict
from typing import List
from typing import cast

import numpy as np
import xarray as xr
from numpy.typing import ArrayLike
from tabulate import tabulate

from glotaran.deprecation import deprecate
from glotaran.io import load_dataset
from glotaran.io import load_parameters
from glotaran.io import load_scheme
from glotaran.io import save_result
from glotaran.model import Model
from glotaran.parameter import ParameterGroup
from glotaran.parameter import ParameterHistory
from glotaran.project.dataclass_helpers import exclude_from_dict_field
from glotaran.project.dataclass_helpers import file_representation_field
from glotaran.project.scheme import Scheme
from glotaran.utils.ipython import MarkdownStr


class IncompleteResultError(Exception):
    """Exception raised if mandatory arguments to create a result are missing.

    Since some mandatory fields of result can be either created from file or by
    passing a class instance, the file and instance initialization aren't allowed
    to both be None at the same time, but each is allowed to be ``None`` by its own.
    """


@dataclass
class Result:
    """The result of a global analysis."""

    number_of_function_evaluations: int
    """The number of function evaluations."""

    success: bool
    """Indicates if the optimization was successful."""

    termination_reason: str
    """The reason (message when) the optimizer terminated"""

    glotaran_version: str
    """The glotaran version used to create the result."""

    free_parameter_labels: list[str]
    """List of labels of the free parameters used in optimization."""

    scheme: Scheme = cast(Scheme, exclude_from_dict_field(None))
    scheme_file: str | None = file_representation_field("scheme", load_scheme, None)

    initial_parameters: ParameterGroup = cast(ParameterGroup, exclude_from_dict_field(None))
    initial_parameters_file: str | None = file_representation_field(
        "initial_parameters", load_parameters, None
    )

    optimized_parameters: ParameterGroup = cast(ParameterGroup, exclude_from_dict_field(None))
    """The optimized parameters, organized in a :class:`ParameterGroup`"""
    optimized_parameters_file: str | None = file_representation_field(
        "optimized_parameters", load_parameters, None
    )

    parameter_history: ParameterHistory = cast(ParameterHistory, exclude_from_dict_field(None))
    """The parameter history."""
    parameter_history_file: str | None = file_representation_field(
        "parameter_history", ParameterHistory.from_csv, None
    )

    data: dict[str, xr.Dataset] = cast(Dict[str, xr.Dataset], exclude_from_dict_field(None))
    """The resulting data as a dictionary of :xarraydoc:`Dataset`.

    Notes
    -----
    The actual content of the data depends on the actual model and can be found in the
    documentation for the model.
    """
    data_files: dict[str, str] | None = file_representation_field("data", load_dataset, None)

    additional_penalty: np.ndarray | None = exclude_from_dict_field(None)
    """A vector with the value for each additional penalty, or None"""

    cost: ArrayLike | None = exclude_from_dict_field(None)
    """The final cost."""

    # The below can be none in case of unsuccessful optimization

    chi_square: float | None = None
    r"""The chi-square of the optimization.

    :math:`\chi^2 = \sum_i^N [{Residual}_i]^2`."""

    covariance_matrix: ArrayLike | None = exclude_from_dict_field(None)
    """Covariance matrix.

    The rows and columns are corresponding to :attr:`free_parameter_labels`."""

    degrees_of_freedom: int | None = None
    """Degrees of freedom in optimization :math:`N - N_{vars}`."""

    jacobian: ArrayLike | list | None = exclude_from_dict_field(None)
    """Modified Jacobian matrix at the solution

    See also: :func:`scipy.optimize.least_squares`
    """
    number_of_data_points: int | None = None
    """Number of data points :math:`N`."""
    number_of_jacobian_evaluations: int | None = None
    """The number of jacobian evaluations."""
    number_of_variables: int | None = None
    """Number of variables in optimization :math:`N_{vars}`"""
    optimality: float | None = None
    reduced_chi_square: float | None = None
    r"""The reduced chi-square of the optimization.

    :math:`\chi^2_{red}= {\chi^2} / {(N - N_{vars})}`.
    """
    root_mean_square_error: float | None = None
    r"""
    The root mean square error the optimization.

    :math:`rms = \sqrt{\chi^2_{red}}`
    """

    def __post_init__(self):
        """Validate fields and cast attributes to correct type."""
        self._check_mandatory_fields()
        if isinstance(self.jacobian, list):
            self.jacobian = np.array(self.jacobian)
            self.covariance_matrix = np.array(self.covariance_matrix)

    def _check_mandatory_fields(self):
        """Check that required fields which can be set from file are not ``None``.

        Raises
        ------
        IncompleteResultError
            If any mandatory field and its file representation is ``None``.
        """
        mandatory_fields = [
            ("scheme", ""),
            ("initial_parameters", ""),
            ("optimized_parameters", ""),
            ("parameter_history", ""),
            ("data", "s"),
        ]
        missing_fields = [
            (mandatory_field, file_post_fix)
            for mandatory_field, file_post_fix in mandatory_fields
            if (
                getattr(self, mandatory_field) is None
                and getattr(self, f"{mandatory_field}_file{file_post_fix}") is None
            )
        ]
        if len(missing_fields) != 0:
            error_message = "Result is missing mandatory fields:\n"
            for missing_field, file_post_fix in missing_fields:
                error_message += (
                    f" - Required filed {missing_field!r} is missing!\n"
                    f"   Set either {missing_field!r} or '{missing_field}_file{file_post_fix}'."
                )
            raise IncompleteResultError(error_message)

    @property
    def model(self) -> Model:
        """Return the model used to fit result.

        Returns
        -------
        Model
            The model instance.
        """
        return self.scheme.model

    def get_scheme(self) -> Scheme:
        """Return a new scheme from the Result object with optimized parameters.

        Returns
        -------
        Scheme
            A new scheme with the parameters set to the optimized values.
            For the dataset weights the (precomputed) weights from the original scheme are used.
        """
        data = {}

        for label, dataset in self.data.items():
            data[label] = dataset.data.to_dataset(name="data")
            if "weight" in dataset:
                data[label]["weight"] = dataset.weight

        return replace(self.scheme, parameters=self.optimized_parameters)

    def markdown(self, with_model: bool = True, base_heading_level: int = 1) -> MarkdownStr:
        """Format the model as a markdown text.

        Parameters
        ----------
        with_model : bool
            If `True`, the model will be printed with initial and optimized parameters filled in.
        base_heading_level : int
            The level of the base heading.

        Returns
        -------
        MarkdownStr : str
            The scheme as markdown string.
        """
        general_table_rows: list[list[Any]] = [
            ["Number of residual evaluation", self.number_of_function_evaluations],
            ["Number of variables", self.number_of_variables],
            ["Number of datapoints", self.number_of_data_points],
            ["Degrees of freedom", self.degrees_of_freedom],
            ["Chi Square", f"{self.chi_square or np.nan:.2e}"],
            ["Reduced Chi Square", f"{self.reduced_chi_square or np.nan:.2e}"],
            ["Root Mean Square Error (RMSE)", f"{self.root_mean_square_error or np.nan:.2e}"],
        ]
        if self.additional_penalty is not None:
            general_table_rows.append(["RMSE additional penalty", self.additional_penalty])

        result_table = tabulate(
            general_table_rows,
            headers=["Optimization Result", ""],
            tablefmt="github",
            disable_numparse=True,
        )
        if len(self.data) > 1:

            RMSE_rows = [
                [
                    f"{index}.{label}:",
                    dataset.weighted_root_mean_square_error,
                    dataset.root_mean_square_error,
                ]
                for index, (label, dataset) in enumerate(self.data.items(), start=1)
            ]

            RMSE_table = tabulate(
                RMSE_rows,
                headers=["RMSE (per dataset)", "weighted", "unweighted"],
                floatfmt=".2e",
                tablefmt="github",
            )

            result_table = f"{result_table}\n\n{RMSE_table}"

        if with_model:
            model_md = self.model.markdown(
                parameters=self.optimized_parameters,
                initial_parameters=self.initial_parameters,
                base_heading_level=base_heading_level,
            )
            result_table = f"{result_table}\n\n{model_md}"

        return MarkdownStr(result_table)

    def _repr_markdown_(self) -> str:
        """Return a markdown representation str.

        Special method used by ``ipython`` to render markdown.

        Returns
        -------
        str
            The scheme as markdown string.
        """
        return str(self.markdown(base_heading_level=3))

    def __str__(self) -> str:
        """Overwrite of ``__str__``."""
        return str(self.markdown(with_model=False))

    def save(self, path: str) -> list[str]:
        """Save the result to given folder.

        Parameters
        ----------
        path : str
            The path to the folder in which to save the result.

        Returns
        -------
        list[str]
            Paths to all the saved files.
        """
        return cast(
            List[str],
            save_result(result_path=path, result=self, format_name="folder", allow_overwrite=True),
        )

    def recreate(self) -> Result:
        """Recrate a result from the initial parameters.

        Returns
        -------
        Result :
            The recreated result.
        """
        from glotaran.analysis.optimize import optimize

        return optimize(self.scheme)

    def verify(self) -> bool:
        """Verify a result.

        Returns
        -------
        bool :
            Weather the recreated result is equal to this result.
        """
        recreated = self.recreate()

        if self.root_mean_square_error != recreated.root_mean_square_error:
            return False

        for label, dataset in self.data.items():
            for attr, array in dataset.items():
                if not np.allclose(array, recreated.data[label][attr]):
                    return False

        return True

    @deprecate(
        deprecated_qual_name_usage="glotaran.project.result.Result.get_dataset(dataset_label)",
        new_qual_name_usage=("glotaran.project.result.Result.data[dataset_label]"),
        to_be_removed_in_version="0.6.0",
        importable_indices=(2, 2),
    )
    def get_dataset(self, dataset_label: str) -> xr.Dataset:
        """Return the result dataset for the given dataset label.

        Warning
        -------
        Deprecated use ``glotaran.project.result.Result.data[dataset_label]``
        instead.


        Parameters
        ----------
        dataset_label : str
            The label of the dataset.

        Returns
        -------
        xr.Dataset :
            The dataset.


        .. # noqa: DAR401
        """
        try:
            return self.data[dataset_label]
        except KeyError:
            raise ValueError(f"Unknown dataset '{dataset_label}'")



"""Functions for calculating conditionally linear parameters and residual with the variable
projection method."""

import typing

import numpy as np
from scipy.linalg import lapack


def residual_variable_projection(
    matrix: np.ndarray, data: np.ndarray
) -> typing.Tuple[np.ndarray, np.ndarray]:
    """Calculates the conditionally linear parameters and residual with the variable projection
    method.

    Parameters
    ----------
    matrix :
        The model matrix.
    data : np.ndarray
        The data to analyze.
    """
    # TODO: Reference Kaufman paper

    # Kaufman Q2 step 3
    qr, tau, _, _ = lapack.dgeqrf(matrix)

    # Kaufman Q2 step 4
    temp, _, _ = lapack.dormqr("L", "T", qr, tau, data, max(1, matrix.shape[1]), overwrite_c=0)

    clp, _ = lapack.dtrtrs(qr, temp)

    for i in range(matrix.shape[1]):
        temp[i] = 0

    # Kaufman Q2 step 5

    residual, _, _ = lapack.dormqr("L", "N", qr, tau, temp, max(1, matrix.shape[1]), overwrite_c=0)
    return clp[: matrix.shape[1]], residual



from IPython.core.formatters import format_display_data

from glotaran.io import load_parameters
from glotaran.parameter.parameter_group import ParameterGroup

PARAMETERS_3C_BASE = """\
irf:
    - ["center", 1.3]
    - ["width", 7.8]
j:
    - ["1", 1, {"vary": False, "non-negative": False}]
"""

PARAMETERS_3C_KINETIC = """\
kinetic:
    - ["1", 300e-3]
    - ["2", 500e-4]
    - ["3", 700e-5]
"""

RENDERED_MARKDOWN = """\
  * __irf__:

    | _Label_   |   _Value_ |   _StdErr_ |   _Min_ |   _Max_ | _Vary_   | _Non-Negative_   | _Expr_   |
    |-----------|-----------|------------|---------|---------|----------|------------------|----------|
    | center    |       1.3 |          0 |    -inf |     inf | True     | False            | None     |
    | width     |       7.8 |          0 |    -inf |     inf | True     | False            | None     |

  * __j__:

    |   _Label_ |   _Value_ |   _StdErr_ |   _Min_ |   _Max_ | _Vary_   | _Non-Negative_   | _Expr_   |
    |-----------|-----------|------------|---------|---------|----------|------------------|----------|
    |         1 |         1 |          0 |    -inf |     inf | False    | False            | None     |

  * __kinetic__:

    |   _Label_ |   _Value_ |   _StdErr_ |   _Min_ |   _Max_ | _Vary_   | _Non-Negative_   | _Expr_   |
    |-----------|-----------|------------|---------|---------|----------|------------------|----------|
    |         1 |     0.3   |          0 |    -inf |     inf | True     | False            | None     |
    |         2 |     0.05  |          0 |    -inf |     inf | True     | False            | None     |
    |         3 |     0.007 |          0 |    -inf |     inf | True     | False            | None     |

"""  # noqa: E501


def test_param_group_markdown_is_order_independent():
    """Markdown output of ParameterGroup.markdown() is independent of initial order"""
    PARAMETERS_3C_INITIAL1 = f"""{PARAMETERS_3C_BASE}\n{PARAMETERS_3C_KINETIC}"""
    PARAMETERS_3C_INITIAL2 = f"""{PARAMETERS_3C_KINETIC}\n{PARAMETERS_3C_BASE}"""

    initial_parameters_ref = ParameterGroup.from_dict(
        {
            "j": [["1", 1, {"vary": False, "non-negative": False}]],
            "kinetic": [
                ["1", 300e-3],
                ["2", 500e-4],
                ["3", 700e-5],
            ],
            "irf": [["center", 1.3], ["width", 7.8]],
        }
    )

    initial_parameters1 = load_parameters(PARAMETERS_3C_INITIAL1, format_name="yml_str")
    initial_parameters2 = load_parameters(PARAMETERS_3C_INITIAL2, format_name="yml_str")

    assert initial_parameters1.markdown() == RENDERED_MARKDOWN
    assert initial_parameters2.markdown() == RENDERED_MARKDOWN
    assert initial_parameters_ref.markdown() == RENDERED_MARKDOWN


def test_param_group_repr():
    """Repr creates code to recreate the object with from_dict."""
    result = ParameterGroup.from_dict({"foo": {"bar": [["1", 1.0], ["2", 2.0], ["3", 3.0]]}})
    result_short = ParameterGroup.from_dict({"foo": {"bar": [1, 2, 3]}})
    expected = "ParameterGroup.from_dict({'foo': {'bar': [['1', 1.0], ['2', 2.0], ['3', 3.0]]}})"

    assert result == result_short
    assert result_short.__repr__() == expected
    assert result.__repr__() == expected
    assert result == eval(result.__repr__())


def test_param_group_repr_from_list():
    """Repr creates code to recreate the object with from_list."""
    result = ParameterGroup.from_list([["1", 2.3], ["2", 3.0]])
    result_short = ParameterGroup.from_list([2.3, 3.0])
    expected = "ParameterGroup.from_list([['1', 2.3], ['2', 3.0]])"

    assert result == result_short
    assert result.__repr__() == expected
    assert result_short.__repr__() == expected
    assert result == eval(result.__repr__())


def test_param_group_ipython_rendering():
    """Autorendering in ipython"""
    param_group = ParameterGroup.from_dict({"foo": {"bar": [["1", 1.0], ["2", 2.0], ["3", 3.0]]}})
    rendered_obj = format_display_data(param_group)[0]

    assert "text/markdown" in rendered_obj
    assert rendered_obj["text/markdown"].startswith("  * __foo__")

    rendered_markdown_return = format_display_data(param_group.markdown())[0]

    assert "text/markdown" in rendered_markdown_return
    assert rendered_markdown_return["text/markdown"].startswith("  * __foo__")



from __future__ import annotations

import pytest
from IPython.core.formatters import format_display_data

from glotaran.analysis.optimize import optimize
from glotaran.analysis.simulation import simulate
from glotaran.analysis.test.models import ThreeDatasetDecay as suite
from glotaran.project import Scheme
from glotaran.project.result import IncompleteResultError
from glotaran.project.result import Result


@pytest.fixture(scope="session")
def dummy_result():
    """Dummy result for testing."""

    wanted_parameters = suite.wanted_parameters
    data = {}
    for i in range(3):
        global_axis = getattr(suite, "global_axis" if i == 0 else f"global_axis{i+1}")
        model_axis = getattr(suite, "model_axis" if i == 0 else f"model_axis{i+1}")

        data[f"dataset{i+1}"] = simulate(
            suite.sim_model,
            f"dataset{i+1}",
            wanted_parameters,
            {"global": global_axis, "model": model_axis},
        )
    scheme = Scheme(
        model=suite.model,
        parameters=suite.initial_parameters,
        data=data,
        maximum_number_function_evaluations=1,
    )

    yield optimize(scheme)


def test_result_ipython_rendering(dummy_result: Result):
    """Autorendering in ipython"""

    rendered_obj = format_display_data(dummy_result)[0]

    assert "text/markdown" in rendered_obj
    assert rendered_obj["text/markdown"].startswith("| Optimization Result")

    rendered_markdown_return = format_display_data(dummy_result.markdown())[0]

    assert "text/markdown" in rendered_markdown_return
    assert rendered_markdown_return["text/markdown"].startswith("| Optimization Result")


def test_result_incomplete_exception(dummy_result: Result):
    """Raise error if required fields are missing."""

    with pytest.raises(IncompleteResultError) as excinfo:
        Result(1, True, "foo", "gta", ["1"])

    for mandatory_field, file_post_fix in [
        ("scheme", ""),
        ("initial_parameters", ""),
        ("optimized_parameters", ""),
        ("parameter_history", ""),
        ("data", "s"),
    ]:
        assert (
            f"Set either '{mandatory_field}' or '{mandatory_field}_file{file_post_fix}'."
            in str(excinfo.value)
        )



"""Test deprecated functionality in 'glotaran.project.schmeme'."""
from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import pytest
import xarray as xr

from glotaran.deprecation.modules.test import deprecation_warning_on_call_test_helper
from glotaran.project.scheme import Scheme
from glotaran.testing.model_generators import SimpleModelGenerator

if TYPE_CHECKING:
    from pathlib import Path


@lru_cache(maxsize=1)
def create_test_args():
    """Objects to initialize a ``Scheme`` for testing."""
    generator = SimpleModelGenerator(
        rates=[501e-3, 202e-4, 105e-5],
        irf={"center": 1.3, "width": 7.8},
        k_matrix="sequential",
    )
    model, parameters = generator.model_and_parameters
    dataset = xr.DataArray([[1, 2, 3]], coords=[("e", [1]), ("c", [1, 2, 3])]).to_dataset(
        name="data"
    )
    return model, parameters, dataset


def test_scheme_from_yaml_file_method(tmp_path: Path):
    """Create Scheme from file."""
    scheme_path = tmp_path / "scheme.yml"

    model_yml_str = """
    megacomplex:
        m1:
            type: decay
            k_matrix: []
    dataset:
        dataset1:
            megacomplex: [m1]
    """
    model_path = tmp_path / "model.yml"
    model_path.write_text(model_yml_str)

    parameter_path = tmp_path / "parameters.yml"
    parameter_path.write_text("[1.0, 67.0]")

    dataset_path = tmp_path / "dataset.nc"
    xr.DataArray([[1, 2, 3]], coords=[("e", [1]), ("c", [1, 2, 3])]).to_dataset(
        name="data"
    ).to_netcdf(dataset_path)

    scheme_path.write_text(
        f"""
        model_file: {model_path}
        parameters_file: {parameter_path}
        maximum_number_function_evaluations: 42
        data_files:
            dataset1: {dataset_path}"""
    )

    _, result = deprecation_warning_on_call_test_helper(
        Scheme.from_yaml_file, args=[str(scheme_path)], raise_exception=True
    )

    assert isinstance(result, Scheme)


def test_scheme_group_tolerance():
    """Argument ``group_tolerance`` raises deprecation and maps to ``clp_link_tolerance``."""
    model, parameters, dataset = create_test_args()

    warnings, result = deprecation_warning_on_call_test_helper(
        Scheme,
        args=(model, parameters, {"dataset": dataset}),
        kwargs={"group_tolerance": 1},
        raise_exception=True,
    )

    assert isinstance(result, Scheme)
    assert result.clp_link_tolerance == 1
    assert "glotaran.project.Scheme(..., clp_link_tolerance=...)" in warnings[0].message.args[0]


@pytest.mark.parametrize(
    "group",
    (True, False),
)
def test_scheme_group(group: bool):
    """Argument ``group`` raises deprecation and maps to ``dataset_groups.default.link_clp``."""
    model, parameters, dataset = create_test_args()

    warnings, result = deprecation_warning_on_call_test_helper(
        Scheme,
        args=(model, parameters, {"dataset": dataset}),
        kwargs={"group": group},
        raise_exception=True,
    )

    assert isinstance(result, Scheme)
    assert result.model.dataset_group_models["default"].link_clp == group
    assert "<model_file>dataset_groups.default.link_clp" in warnings[0].message.args[0]


@pytest.mark.parametrize(
    "non_negative_least_squares, expected",
    ((True, "non_negative_least_squares"), (False, "variable_projection")),
)
def test_scheme_non_negative_least_squares(non_negative_least_squares: bool, expected: str):
    """Argument ``non_negative_least_squares`` raises deprecation and maps to
    ``dataset_groups.default.residual_function``.
    """
    model, parameters, dataset = create_test_args()

    warnings, result = deprecation_warning_on_call_test_helper(
        Scheme,
        args=(model, parameters, {"dataset": dataset}),
        kwargs={"non_negative_least_squares": non_negative_least_squares},
        raise_exception=True,
    )

    assert isinstance(result, Scheme)
    assert result.model.dataset_group_models["default"].residual_function == expected
    assert "<model_file>dataset_groups.default.residual_function" in warnings[0].message.args[0]
